import logging
mpl_logger = logging.getLogger('matplotlib')
mpl_logger.setLevel(logging.WARNING)

import unittest
import pygsti
import numpy as np
from pygsti.modelpacks import smq1Q_XYI
from pygsti.circuits import Circuit
from pygsti.baseobjs import Label

from ..testutils import BaseTestCase


class MyTimeDependentIdle(pygsti.modelmembers.operations.DenseOperator):
    """And idle that depolarizes over time with a parameterized rate"""
    def __init__(self, initial_depol_rate):
        #initialize with no noise
        self.need_time = True # maybe torep() won't work unless this is False?
        super(MyTimeDependentIdle,self).__init__(np.identity(4,'d'), 'pp', "densitymx") # this is *super*-operator, so "densitymx"
        self.from_vector([initial_depol_rate])
        self.set_time(0.0)

    @property
    def num_params(self):
        return 1 # we have two parameters

    def to_vector(self):
        return np.array([self.depol_rate],'d') #our parameter vector

    def from_vector(self, v, close=False, dirty_value=True):
        #initialize from parameter vector v
        self.depol_rate = v[0]
        self.need_time = True
        self.dirty = dirty_value

    def set_time(self,t):
        a = 1.0-min(self.depol_rate*t,1.0)
        self.need_time = False

        # .base is a member of DenseOperator and is a numpy array that is
        # the dense Pauli transfer matrix of this operator
        self._ptr[:,:] = np.array([[1,   0,   0,   0],
                                   [0,   a,   0,   0],
                                   [0,   0,   a,   0],
                                   [0,   0,   0,   a]],'d')
        self._ptr_has_changed()

    def transform(self, S):
        # Update self with inverse(S) * self * S (used in gauge optimization)
        raise NotImplementedError("MyTimeDependentIdle cannot be transformed!")


class TimeDependentTestCase(BaseTestCase):

    def setUp(self):
        super(TimeDependentTestCase, self).setUp()

    def test_time_dependent_datagen(self):
        mdl = smq1Q_XYI.target_model("full TP")
        mdl.sim = 'map'
        mdl.operations['Gi',0] = MyTimeDependentIdle(1.0)

        #Create a time-dependent dataset (simulation of time-dependent model):
        circuits = smq1Q_XYI.prep_fiducials() + [Circuit([Label('Gi',0)], line_labels=(0,)), 
                                                Circuit([Label('Gi',0), Label('Gxpi2',0), Label('Gi',0), Label('Gxpi2',0)], line_labels=(0,))] 
                                                # just pick some circuits
        ds = pygsti.data.simulate_data(mdl, circuits, num_samples=100,
                                       sample_error='none', seed=1234, times=[0,0.1,0.2])

        self.assertArraysEqual(ds[Circuit([Label('Gi',0)], line_labels=(0,))].time, np.array([0.,  0.,  0.1, 0.1, 0.2, 0.2]))
        self.assertArraysEqual(ds[Circuit([Label('Gi',0)], line_labels=(0,))].reps, np.array([100.,   0.,  95.,   5.,  90.,  10.]))
        self.assertArraysEqual(ds[Circuit([Label('Gi',0)], line_labels=(0,))].outcomes, [('0',), ('1',), ('0',), ('1',), ('0',), ('1',)])

        # sparse data
        ds2 = pygsti.data.simulate_data(mdl, circuits, num_samples=100,
                                        sample_error='none', seed=1234, times=[0,0.1,0.2],
                                        record_zero_counts=False)
        self.assertArraysEqual(ds2[Circuit([Label('Gi',0)], line_labels=(0,))].time, np.array([0.,  0.1, 0.1, 0.2, 0.2]))
        self.assertArraysEqual(ds2[Circuit([Label('Gi',0)], line_labels=(0,))].reps, np.array([100.,  95.,   5.,  90.,  10.]))
        self.assertArraysEqual(ds2[Circuit([Label('Gi',0)], line_labels=(0,))].outcomes, [('0',), ('0',), ('1',), ('0',), ('1',)])

    def test_time_dependent_gst_staticdata(self):
        
        #run GST in a time-dependent mode:
        prep_fiducials, meas_fiducials = smq1Q_XYI.prep_fiducials()[0:4], smq1Q_XYI.meas_fiducials()[0:3]
        germs = smq1Q_XYI.germs(lite=True)
        germs[0] = Circuit([Label('Gi',0)], line_labels=(0,))
        maxLengths = [1, 2]

        target_model = smq1Q_XYI.target_model("full TP")
        target_model.sim = "map"
        
        del target_model.operations[Label(())]
        target_model.operations['Gi',0] = np.eye(4)
        
        mdl_datagen = target_model.depolarize(op_noise=0.05, spam_noise=0.01)
        edesign = pygsti.protocols.StandardGSTDesign(target_model.create_processor_spec(), prep_fiducials,
                                                     meas_fiducials, germs, maxLengths)

        # *sparse*, time-independent data
        ds = pygsti.data.simulate_data(mdl_datagen, edesign.all_circuits_needing_data, num_samples=1000,
                                       sample_error="binomial", seed=1234, times=[0],
                                       record_zero_counts=False)
        data = pygsti.protocols.ProtocolData(edesign, ds)
        target_model.sim = pygsti.forwardsims.MapForwardSimulator(max_cache_size=0)  # No caching allowed for time-dependent calcs
        self.assertEqual(ds.degrees_of_freedom(aggregate_times=False), 57)
        
        builders = pygsti.protocols.GSTObjFnBuilders([pygsti.objectivefns.TimeDependentPoissonPicLogLFunction.builder()], [])
        gst = pygsti.protocols.GateSetTomography(target_model, gaugeopt_suite=None,
                                                 objfn_builders=builders,
                                                 optimizer={'maxiter':2,'tol': 1e-4})
        results = gst.run(data)

        # Normal GST used as a check - should get same answer since data is time-independent
        #We aren't actually doing this comparison atm (relevant tests are commented out) so no point
        #doing the computation. For some reason this fit also took very long to run, which is strange (I don't see
        #any reason why it would)
        #results2 = pygsti.run_long_sequence_gst(ds, target_model, prep_fiducials, meas_fiducials,
        #                                        germs, maxLengths, verbosity=3,
        #                                        advanced_options={'starting_point': 'target',
        #                                                          'always_perform_mle': True,
        #                                                          'only_perform_mle': True}, gauge_opt_params=False)
        
        #These check FAIL on some TravisCI machines for an unknown reason (but passes on Eriks machines) -- figure out why this is in FUTURE.
        #Check that "timeDependent=True" mode matches behavior or "timeDependent=False" mode when model and data are time-independent.
        #self.assertAlmostEqual(pygsti.tools.chi2(results.estimates['default'].models['iteration estimates'][0], results.dataset, results.circuit_lists['iteration'][0]),
        #                       pygsti.tools.chi2(results2.estimates['default'].models['iteration estimates'][0], results2.dataset, results2.circuit_lists['iteration'][0]),
        #                       places=0)
        #self.assertAlmostEqual(pygsti.tools.chi2(results.estimates['default'].models['iteration estimates'][1], results.dataset, results.circuit_lists['iteration'][1]),
        #                       pygsti.tools.chi2(results2.estimates['default'].models['iteration estimates'][1], results2.dataset, results2.circuit_lists['iteration'][1]),
        #                       places=0)
        #self.assertAlmostEqual(pygsti.tools.two_delta_logl(results.estimates['default'].models['final iteration estimate'], results.dataset),
        #                       pygsti.tools.two_delta_logl(results2.estimates['default'].models['final iteration estimate'], results2.dataset),
        #                       places=0)

    def test_time_dependent_gst(self):
        #run GST in a time-dependent mode:
        #use minimally informationally complete set
        prep_fiducials, meas_fiducials = smq1Q_XYI.prep_fiducials()[0:4], smq1Q_XYI.meas_fiducials()[0:3]
        germs = smq1Q_XYI.germs(lite=True)
        germs[0] = Circuit([Label('Gi',0)], line_labels=(0,))
        maxLengths = [1, 2]

        target_model = smq1Q_XYI.target_model("full TP")
        target_model.sim = 'map'
        del target_model.operations[Label(())]
        mdl_datagen = target_model.depolarize(op_noise=0.05, spam_noise=0.01)
        mdl_datagen.operations['Gi',0] = MyTimeDependentIdle(1.0)
        edesign = pygsti.protocols.StandardGSTDesign(target_model.create_processor_spec(), prep_fiducials,
                                                     meas_fiducials, germs, maxLengths)

        # *sparse*, time-independent data
        ds = pygsti.data.simulate_data(mdl_datagen, edesign.all_circuits_needing_data, num_samples=2000,
                                       sample_error="binomial", seed=1234, times=[0, 0.2],
                                       record_zero_counts=False)
        self.assertEqual(ds.degrees_of_freedom(aggregate_times=False), 114)

        target_model.operations['Gi',0] = MyTimeDependentIdle(0)  # start assuming no time dependent decay
        target_model.sim = pygsti.forwardsims.MapForwardSimulator(max_cache_size=0)  # No caching allowed for time-dependent calcs

        builders = pygsti.protocols.GSTObjFnBuilders([pygsti.objectivefns.TimeDependentPoissonPicLogLFunction.builder()], [])
        gst = pygsti.protocols.GateSetTomography(target_model, gaugeopt_suite=None,
                                                 objfn_builders=builders, optimizer={'maxiter':10,'tol': 1e-4})
        data = pygsti.protocols.ProtocolData(edesign, ds)
        results = gst.run(data)

        #we should recover the 1.0 decay we put into mdl_datagen['Gi']:
        final_mdl = results.estimates['GateSetTomography'].models['final iteration estimate']
        print("Final decay rate = ", final_mdl.operations['Gi',0].to_vector())
        #self.assertAlmostEqual(final_mdl.operations['Gi',0].to_vector()[0], 1.0, places=1)
        self.assertAlmostEqual(final_mdl.operations['Gi',0].to_vector()[0], 1.0, delta=0.1) # weaker b/c of unknown TravisCI issues

if __name__ == "__main__":
    unittest.main(verbosity=2)
